@@ -51,6 +51,7 @@
   (:require [clojure.java.jdbc :as sql]
             [clojure.tools.logging :as log]
             [clojure.string :as string]
+            [com.puppetlabs.puppetdb.scf.migration-legacy :as legacy]
             [com.puppetlabs.puppetdb.scf.storage :as scf-store]
             [com.puppetlabs.cheshire :as json]
             [puppetlabs.kitchensink.core :as kitchensink]
@@ -723,7 +724,7 @@
                              first
                              :name)]
         (when-not (empty? facts)
-          (scf-store/add-facts!
+          (legacy/add-facts-27!
             {:name (str certname)
              :values facts
              :timestamp timestamp
@@ -883,6 +884,99 @@
     (sql/do-commands
      "DROP INDEX fact_values_string_trgm")))
 
+(defn- lift-fact-paths-into-facts
+  "Pairs paths and values directly in facts, i.e. change facts from (id
+  value) to (id path value)."
+  []
+  (sql/do-commands
+   ;; Build complete facts table as of migration 28.
+
+   "CREATE TABLE facts_unique_transform
+      (factset_id bigint NOT NULL,
+       fact_path text NOT NULL,
+       fact_value_hash varchar(40) NOT NULL)"
+
+   "INSERT INTO facts_unique_transform (factset_id, fact_path, fact_value_hash)
+      SELECT f.factset_id, fp.path, fv.value_hash
+        FROM facts f
+        INNER JOIN fact_values fv on f.fact_value_id = fv.id
+        INNER JOIN fact_paths fp on fv.path_id = fp.id"
+
+   "DROP TABLE facts"
+
+   ;; We need these indexes for the upcoming deletions.
+   ;; A deletion like this doesn't need the index, but it only works
+   ;; with PostgreSQL:
+   ;;   DELETE FROM fact_values t1 USING fact_values t2
+   ;;     WHERE t1.value_hash = t2.value_hash AND t1.id > t2.id
+   "CREATE INDEX fact_path_path_idx ON fact_paths(path)"
+   "CREATE INDEX fact_values_value_hash_idx ON fact_values(value_hash)"
+
+   ;; Remove all the orphaned duplicates (all but the row in each set
+   ;; with min-id).
+   "DELETE FROM fact_paths t1
+      WHERE t1.id <> (SELECT MIN(t2.id) FROM fact_paths t2
+                        WHERE t1.path = t2.path)"
+   "DELETE FROM fact_values t1
+      WHERE t1.id <> (SELECT MIN(t2.id) FROM fact_values t2
+                        WHERE t1.value_hash = t2.value_hash)"
+
+   ;; Q: Is this right, or do we just keep the (redundant) indexes?
+   "DROP INDEX fact_path_path_idx"
+   "DROP INDEX fact_values_value_hash_idx"
+   "ALTER TABLE fact_paths
+      ADD CONSTRAINT fact_paths_path_key UNIQUE (path)"
+   "ALTER TABLE fact_values
+      ADD CONSTRAINT fact_values_value_hash_key UNIQUE (value_hash)"
+
+   "CREATE TABLE facts_transform
+      (factset_id bigint NOT NULL,
+       fact_path_id bigint NOT NULL,
+       fact_value_id bigint NOT NULL)"
+
+   ;; Patch up facts refrences to refer to the min id path/value
+   ;; wherever there's more than one option.
+   "INSERT INTO facts_transform (factset_id, fact_path_id, fact_value_id)
+      SELECT f.factset_id, fp.id, fv.id
+        FROM facts_unique_transform f
+        INNER JOIN fact_paths fp on f.fact_path = fp.path
+        INNER JOIN fact_values fv on f.fact_value_hash = fv.value_hash;"
+
+   "DROP TABLE facts_unique_transform"
+
+   "ALTER TABLE facts_transform
+      ADD CONSTRAINT facts_factset_id_fact_path_id_fact_key
+        UNIQUE (factset_id, fact_path_id)"
+
+   "ALTER TABLE facts_transform
+      ADD CONSTRAINT factset_id_fk
+        FOREIGN KEY (factset_id) REFERENCES factsets(id)
+        ON UPDATE CASCADE
+        ON DELETE CASCADE"
+
+   "ALTER TABLE facts_transform
+         ADD CONSTRAINT fact_path_id_fk
+           FOREIGN KEY (fact_path_id) REFERENCES fact_paths(id)"
+   "ALTER TABLE facts_transform
+         ADD CONSTRAINT fact_value_id_fk
+           FOREIGN KEY (fact_value_id) REFERENCES fact_values(id)"
+
+   "ALTER TABLE facts_transform RENAME TO facts"
+   "CREATE INDEX facts_fact_path_id_idx ON facts(fact_path_id)"
+   "CREATE INDEX facts_fact_value_id_idx ON facts(fact_value_id)"
+
+   (if (scf-utils/postgres?) "ANALYZE facts" "SELECT 1")
+
+   ;; These are for the more pedantic HSQLDB.
+   "ALTER TABLE fact_paths DROP CONSTRAINT fact_paths_path_type_id_key"
+   "ALTER TABLE fact_paths DROP CONSTRAINT fact_paths_value_type_id"
+   "DROP INDEX fact_paths_value_type_id"
+   "ALTER TABLE fact_values DROP CONSTRAINT fact_values_path_id_value_key"
+   "ALTER TABLE fact_values DROP CONSTRAINT fact_values_path_id_fk"
+
+   "ALTER TABLE fact_paths DROP COLUMN value_type_id"
+   "ALTER TABLE fact_values DROP COLUMN path_id"))
+
 ;; The available migrations, as a map from migration version to migration function.
 (def migrations
   {1 initialize-store
@@ -911,7 +1005,8 @@
    24 add-producer-timestamps
    25 structured-facts
    26 structured-facts-deferrable-constraints
-   27 switch-value-string-index-to-gin})
+   27 switch-value-string-index-to-gin
+   28 lift-fact-paths-into-facts})
 
 (def desired-schema-version (apply max (keys migrations)))
 
@@ -931,12 +1026,14 @@
   {:post  [(sorted? %)
            (set? %)
            (apply < 0 %)]}
-  (try
-    (let [query   "SELECT version FROM schema_migrations ORDER BY version"
-          results (sql/transaction (query-to-vec query))]
-      (apply sorted-set (map :version results)))
-    (catch java.sql.SQLException e
-      (sorted-set))))
+  (sql/transaction
+   (if-not (some #(= (string/lower-case %) "schema_migrations")
+                 (scf-utils/sql-current-connection-table-names))
+     (sorted-set)
+     (apply sorted-set
+            (map :version
+                 (query-to-vec
+                  "SELECT version FROM schema_migrations ORDER BY version"))))))
 
 (defn pending-migrations
   "Returns a collection of pending migrations, ordered from oldest to latest."
